/*
 * Copyright (c) 2007, 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Jan 27, 2008
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */

package pl.edu.icm.uvos.webapp.actions.idform;

import org.apache.log4j.Logger;
import org.bouncycastle.asn1.x509.X509Name;
import org.bouncycastle.jce.PKCS10CertificationRequest;
import org.springframework.validation.Errors;
import org.springframework.webflow.action.FormAction;
import org.springframework.webflow.execution.RequestContext;

import eu.unicore.security.CertificateUtils;

import java.security.InvalidAlgorithmParameterException;
import java.security.cert.CertPath;
import java.security.cert.CertPathValidator;
import java.security.cert.CertPathValidatorException;
import java.security.cert.CertificateExpiredException;
import java.security.cert.CertificateFactory;
import java.security.cert.CertificateNotYetValidException;
import java.security.cert.PKIXParameters;
import java.security.cert.TrustAnchor;
import java.security.cert.X509Certificate;
import java.util.Collections;
import java.util.Hashtable;
import java.util.List;
import java.util.Set;

import javax.security.auth.x500.X500Principal;

import pl.edu.icm.unicore.uvos.api.Identity;
import pl.edu.icm.unicore.uvos.api.IdentityType;
import pl.edu.icm.unicore.uvos.api.exceptions.InvalidValueException;
import pl.edu.icm.unicore.uvos.idhelpers.X509IdentityHelper;
import pl.edu.icm.uvos.webapp.actions.appform.AppFormData;
import pl.edu.icm.uvos.webapp.data.TrustStore;



/**
 * @author K. Benedyczak
 */
public class IdentityFormAction extends FormAction
{
	private static final Logger log = Logger.getLogger(IdentityFormAction.class);
	private Set<TrustAnchor> taSet;
	
	public IdentityFormAction(TrustStore ts)
	{
		taSet = ts.getTrustAnchors();
	}
	
	@Override
	protected void doValidate(RequestContext context, Object formObject,
			Errors errors) throws java.lang.Exception
	{
		super.doValidate(context, formObject, errors);
		IdFormData dat = (IdFormData) formObject;
		
		AppFormData dat0 = (AppFormData) 
			context.getFlowScope().get("app1FormData");
		Boolean disableCertValidation = 
			(Boolean)context.getFlowScope().get("disableCertificateCheck");
		if (disableCertValidation == null)
			disableCertValidation = false;
		
		if (dat0.getIsCSR() != null && dat0.getIsCSR().size() > 0)
		{
			validateCSR(dat, errors);
		} else if (dat.getIdentityType().equals(IdentityType.DN.getName()))
			validateDN(dat, errors);
		else if (dat.getIdentityType().equals(IdentityType.EMAIL.getName()))
			validateEmail(dat, errors);
		else if (dat.getIdentityType().equals(IdentityType.X509CERT.getName()))
			validateCert(dat, errors, disableCertValidation);
			
		if (dat.getCaCertError() != null)
		{
			errors.rejectValue("caCertError", "caCertError", 
					dat.getCaCertError());
		} else if (dat.getCaCert() != null)
		{
			parsePEM(dat.getCaCert(), errors, "caCertError");
		}
	}
	
	private void validateDN(IdFormData dat, Errors errors)
	{
		if (dat.getIdDN() == null || dat.getIdDN().length() == 0)
		{
			if (dat.getIdCertError() != null)
			{
				errors.rejectValue("idCertError", "idCertError", 
					dat.getIdCertError());
				errors.rejectValue("idDN", "idDN", 
					"You must provide your DN (either directly or via certificate)");
				return;
			}
			X509Certificate cert = parsePEM(dat.getIdCert(), errors, "idCertError");
			if (cert == null)
				return;
			dat.setIdDN(cert.getSubjectX500Principal().
					getName());
		} else
		{
			try
			{
				X500Principal x500 = new X500Principal(dat.getIdDN());
				dat.setIdDN(x500.getName());
			} catch (Exception e)
			{
				errors.rejectValue("idDN", "idDN", "Entered DN is incorrect");
			}
		}
	}
	
	private void validateEmail(IdFormData dat, Errors errors)
	{
		String email = dat.getEmail();
		String p1 = dat.getPasswd1();
		String p2 = dat.getPasswd2();
		
		if (p1 == null)
			errors.rejectValue("passwd1", "passwd1", 
					"Password can't be empty");
		if (p2 == null)
			errors.rejectValue("passwd2", "passwd2", 
					"Password can't be empty");
		if (email == null)
			errors.rejectValue("email", "email", 
					"Email address can't be empty");
		if (p1 == null || p2 == null || email == null)
			return;
		
		if (!p1.equals(p2))
			errors.rejectValue("passwd1", "passwd1", 
				"Passwords do not match");
		
		if (p1.length() < 4)
			errors.rejectValue("passwd1", "passwd1", 
				"Password is too short");
		int mi = email.indexOf("@");
		if (mi == -1 || mi == 0 || mi == email.length() - 1)
			errors.rejectValue("email", "email", 
				"Your email address is incorrect");
	}
	
	private void validateCert(IdFormData dat, Errors errors, boolean disableFullValidation)
	{
		if (dat.getIdCertError() != null)
		{
			errors.rejectValue("idCertError", "idCertError", 
					dat.getIdCertError());
			return;
		}
		X509Certificate c = parsePEM(dat.getIdCert(), errors, "idCertError");
		if (c == null)
			return;
		if (!disableFullValidation)
		{
			try
			{
				checkCert(c);
			} catch (CertPathValidatorException ce)
			{
				errors.rejectValue("idCertError", "idCertError", 
						"Certificate is not valid. Common problems include expired " +
						"certificate or issued by an authority which is not honored " +
						"by this VO server. Details: " + ce.getMessage());
				return;
			} catch (InvalidAlgorithmParameterException alge) 
			{
				errors.rejectValue("idCertError", "idCertError", 
						"Certificate uses an algorithm or " +
						"extension which is not supported by this server. Details: " + alge.getMessage());
				return;
			} catch (Exception e)
			{
				log.error(e);
				errors.rejectValue("idCertError", "idCertError", 
					e.getMessage() == null ? e.toString() : e.getMessage());
				return;
			}
			
			try
			{
				CertificateUtils.verifyCertificate(c, true, false);
			} catch (CertificateExpiredException e)
			{
				errors.rejectValue("idCertError", "idCertError", 
						"Certificate is expired: " + e.getMessage());
				return;
			} catch (CertificateNotYetValidException e)
			{
				errors.rejectValue("idCertError", "idCertError", 
						"Certificate is not yet valid: " + e.getMessage());
				return;
			}
		}
		
		dat.setIdDN(c.getSubjectX500Principal().getName(
				X500Principal.CANONICAL));
	}
	
	private void checkCert(X509Certificate cert) throws Exception
	{
		PKIXParameters params = new PKIXParameters(taSet);
		params.setRevocationEnabled(false);
		CertPathValidator validator = CertPathValidator.getInstance("PKIX");
		
		CertificateFactory fact = CertificateFactory.getInstance("X.509");
		
		List<X509Certificate> cc = Collections.singletonList(cert);
		CertPath cp = fact.generateCertPath(cc);
		
		validator.validate(cp, params);
	}
	
	private void validateCSR(IdFormData dat, Errors errors)
	{
		if (dat.getIdCertError() != null)
		{
			errors.rejectValue("idCertError", "idCertError", 
					dat.getIdCertError());
			return;
		}
		String t = dat.getIdCert();
		try
		{
			PKCS10CertificationRequest csr = X509IdentityHelper.parseCSR(t);
			X509Name subject = csr.getCertificationRequestInfo().getSubject(); 
			String dn = subject.toString(true, new Hashtable<Object, Object>());
			dat.setIdDN(new X500Principal(dn).getName(X500Principal.CANONICAL));
		} catch (Exception e)
		{
			errors.rejectValue("idCertError", "idCertError", 
					"Invalid CSR (must be in PEM format).");
		}
	}
	
	private X509Certificate parsePEM(String pem, Errors errors, String id)
	{
		Identity tmp = new Identity(IdentityType.X509CERT, pem);
		try
		{
			X509IdentityHelper helper = new X509IdentityHelper(tmp);
			return helper.getCertificate();
		} catch (InvalidValueException e)
		{
			errors.rejectValue(id, id, e.getMessage());
			return null;
		}
	}
}






